/*
 * moonlight/core.h: Standalone essentials for everyday coding.
 *
 * 06/30/2020: This header is deprecated.  It contained functionality for many different
 * things that were considered "core" at the time, but it grew to be too large so
 * it was split among mulitple files.  The files who's functionality used to be
 * part of the core have been included here for backward compatibility.
 *
 * Author: Lain Supe (lainproliant)
 * Date: Thursday, Dec 21 2017
 */
#ifndef __MOONLIGHT_CORE_H
#define __MOONLIGHT_CORE_H

#include "moonlight/string.h"
#include "moonlight/exceptions.h"
#include "moonlight/debug.h"
#include "moonlight/make.h"
#include "moonlight/variadic.h"
#include "moonlight/system.h"
#include "moonlight/hash.h"
#include "moonlight/maps.h"
#include "moonlight/mmap.h"
#include "moonlight/collect.h"
#include "moonlight/file.h"

#endif /* __MOONLIGHT_CORE_H */
